import dataclasses
import io
import json
from unittest.mock import patch

import pytest

from jhub_apps.hub_client.hub_client import HubClient
from jhub_apps.spawner.types import FRAMEWORKS
from jhub_apps.tests.constants import MOCK_USER


def mock_user_options():
    user_options = {
        "jhub_app": True,
        "display_name": "Panel App",
        "description": "description",
        "thumbnail": "",
        "filepath": "",
        "framework": "panel",
        "custom_command": "",
        "conda_env": "",
        "profile": "",
        "public": False
    }
    return user_options


@patch.object(HubClient, "get_users")
def test_api_get_server(get_users, client):
    server_data = {"panel-app": {}}
    get_users_response = [
        {'name': 'aktech', 'servers': server_data}
    ]
    get_users.return_value = get_users_response
    response = client.get("/server/panel-app")
    get_users.assert_called_once_with()
    assert response.status_code == 200
    assert response.json() == server_data["panel-app"]


@patch.object(HubClient, "create_server")
def test_api_create_server(create_server, client):
    from jhub_apps.service.models import UserOptions
    create_server_response = {"user": "aktech"}
    create_server.return_value = create_server_response
    user_options = mock_user_options()
    thumbnail = b"contents of thumbnail"
    in_memory_file = io.BytesIO(thumbnail)
    response = client.post(
        "/server",
        data={'data': json.dumps({"servername": "panel-app", "user_options": user_options})},
        files={'thumbnail': ('image.jpeg', in_memory_file)}
    )
    final_user_options = UserOptions(**user_options)
    final_user_options.thumbnail = "data:image/jpeg;base64,Y29udGVudHMgb2YgdGh1bWJuYWls"
    create_server.assert_called_once_with(
        username=MOCK_USER.name,
        servername="panel-app",
        user_options=final_user_options,
    )
    assert response.status_code == 200
    assert response.json() == create_server_response


@patch.object(HubClient, "start_server")
def test_api_start_server(create_server, client):
    start_server_response = {"user": "aktech"}
    create_server.return_value = start_server_response
    server_name = "server-name"
    response = client.post(
        f"/server/{server_name}",
    )
    create_server.assert_called_once_with(
        username=MOCK_USER.name,
        servername=server_name,
    )
    assert response.status_code == 200
    assert response.json() == start_server_response


@patch.object(HubClient, "start_server")
def test_api_start_server_404(start_server, client):
    start_server_response = None
    start_server.return_value = start_server_response
    server_name = "server-name"
    response = client.post(
        f"/server/{server_name}",
    )
    start_server.assert_called_once_with(
        username=MOCK_USER.name,
        servername=server_name,
    )
    assert response.status_code == 404
    assert response.json() == {"detail": "server 'server-name' not found"}


@pytest.mark.parametrize("name,remove", [
    ('delete', True,),
    ('stop', False,),
])
@patch.object(HubClient, "delete_server")
def test_api_delete_server(delete_server, name, remove, client):
    create_server_response = {"user": "aktech"}
    delete_server.return_value = create_server_response
    response = client.delete("/server/panel-app", params={"remove": remove})
    delete_server.assert_called_once_with(
        MOCK_USER.name,
        server_name="panel-app",
        remove=remove
    )
    assert response.status_code == 200
    assert response.json() == create_server_response


@patch.object(HubClient, "edit_server")
def test_api_update_server(edit_server, client):
    from jhub_apps.service.models import UserOptions

    create_server_response = {"user": "aktech"}
    edit_server.return_value = create_server_response
    user_options = mock_user_options()
    thumbnail = b"contents of thumbnail"
    in_memory_file = io.BytesIO(thumbnail)
    response = client.put(
        "/server/panel-app",
        data={'data': json.dumps({"servername": "panel-app", "user_options": user_options})},
        files={'thumbnail': ('image.jpeg', in_memory_file)}
    )
    final_user_options = UserOptions(**user_options)
    final_user_options.thumbnail = "data:image/jpeg;base64,Y29udGVudHMgb2YgdGh1bWJuYWls"
    edit_server.assert_called_once_with(
        username=MOCK_USER.name,
        servername="panel-app",
        user_options=final_user_options,
    )
    assert response.status_code == 200
    assert response.json() == create_server_response


def test_api_frameworks(client):
    response = client.get(
        "/frameworks",
    )
    frameworks = []
    for framework in FRAMEWORKS:
        frameworks.append(dataclasses.asdict(framework))
    assert response.json() == frameworks


def test_api_status(client):
    response = client.get(
        "/status",
    )
    assert response.status_code == 200
    rjson = response.json()
    assert rjson["status"] == "ok"
    assert "version" in rjson



from urllib.parse import urlparse


def get_origin_host(bind_url):
    parsed_url = urlparse(bind_url)
    if "0.0.0.0" in parsed_url.netloc:
        # Hack: Useful for local development when using docker
        # Maybe take it from the user via JAppsConfig?
        return "127.0.0.1:8000"
    return parsed_url.netloc



from pathlib import Path

from jhub_apps.config_utils import JAppsConfig  # noqa: F401

HERE = Path(__file__).parent.resolve()

TEMPLATE_PATH = HERE.joinpath("templates")

theme_template_paths = [TEMPLATE_PATH]



import os

import pytest

from fastapi.testclient import TestClient

from jhub_apps.tests.constants import MOCK_USER


@pytest.fixture
def client():
    os.environ["JUPYTERHUB_API_URL"] = "/"
    os.environ["JUPYTERHUB_API_TOKEN"] = "token"
    os.environ["PUBLIC_HOST"] = "/"
    os.environ["JUPYTERHUB_CLIENT_ID"] = "test-client-id"
    os.environ["JUPYTERHUB_OAUTH_CALLBACK_URL"] = "/"
    from jhub_apps.service import app
    from jhub_apps.service.security import get_current_user

    async def mock_get_user_name():
        return MOCK_USER

    app.dependency_overrides[get_current_user] = mock_get_user_name
    return TestClient(app)



from traitlets import Unicode, Union, List, Callable, Integer
from traitlets.config import SingletonConfigurable, Enum


class JAppsConfig(SingletonConfigurable):
    apps_auth_type = Enum(
        values=["oauth", "none"],
        default_value="oauth",
        help="Authentication for deployed apps, either",
    ).tag(config=True)

    python_exec = Unicode(
        "python", help="Python executable to use for running all the commands"
    ).tag(config=True)

    app_title = Unicode(
        "JHub Apps Launcher",
        help="Title to display on the Home Page of JHub Apps Launcher",
    ).tag(config=True)

    app_icon = Unicode(
        "https://jupyter.org/assets/homepage/main-logo.svg",
        help="Icon to display on the Home Page of JHub Apps Launcher",
    ).tag(config=True)

    conda_envs = Union(
        [List(trait=Unicode, default_value=[], minlen=0), Callable()],
        help="""
        A list of conda environment names for the app creator to display as dropdown to select.
        Default value is the empty list. Also accepts conda_envs to be a callable function
        which is evaluated for each render of the create app page.
        """,
    ).tag(config=True)

    # This is to make sure that JupyterHub config is accessible inside services
    # There is no way to get JupyterHub config inside services otherwise.
    jupyterhub_config_path = Unicode(
        "jupyterhub_config.py",
        help="Path to JupyterHub config file.",
    ).tag(config=True)

    hub_host = Unicode(
        "127.0.0.1",
        help="Hub Host name, in k8s environment it would be the container name, e.g. 'hub'",
    ).tag(config=True)

    service_workers = Integer(
        2,
        help="The number of workers to create for the JHub Apps FastAPI service",
    ).tag(config=True)



"""We can use this to test the bokeh_root_cmd"""
import panel as pn


css = """
body {
    font-family: Mukta, sans-serif;
}

.center-text {
    text-align: center;
}
"""
pn.extension(sizing_mode="stretch_width", raw_css=[css])


def test_panel_app():
    """Returns a Panel test app that has been marked `.servable()`

    Returns:
        pn.Column: A Column based Panel app
    """
    slider = pn.widgets.FloatSlider(name="Slider")
    return pn.template.FastListTemplate(
        title="Panel Test App", sidebar=[slider], main=[slider.param.value]
    ).servable()


if __name__.startswith("bokeh"):
    test_panel_app()



from jhub_apps.spawner.command import TString, Command


def test_tstring():
    filepath = "foo/bar"
    assert (
        TString("random $filepath").replace(filepath=filepath) == f"random {filepath}"
    )


def test_cmd_templating():
    cmd = Command(
        args=[TString("alpha $abc"), "beta", TString("$efg"), TString("$foo $bar")]
    )
    s_args = cmd.get_substituted_args(abc="abc_", efg="_efg", foo="foo_", bar="_bar")
    assert s_args == ["alpha abc_", "beta", "_efg", "foo_ _bar"]



import os

import httpx


# a minimal alternative to using HubOAuth class
def get_client():
    base_url = os.environ["JUPYTERHUB_API_URL"]
    token = os.environ["JUPYTERHUB_API_TOKEN"]
    headers = {"Authorization": "Bearer %s" % token}
    return httpx.AsyncClient(base_url=base_url, headers=headers)



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

# Configuration file for JupyterHub
import os

c = get_config()  # noqa: F821

# We rely on environment variables to configure JupyterHub so that we
# avoid having to rebuild the JupyterHub container every time we change a
# configuration parameter.

# Spawn single-user servers as Docker containers
c.JupyterHub.spawner_class = "dockerspawner.DockerSpawner"

# Spawn containers from this image
c.DockerSpawner.image = os.environ["DOCKER_NOTEBOOK_IMAGE"]

# Connect containers to this Docker network
network_name = os.environ["DOCKER_NETWORK_NAME"]
c.DockerSpawner.use_internal_ip = True
c.DockerSpawner.network_name = network_name

# Explicitly set notebook directory because we'll be mounting a volume to it.
# Most `jupyter/docker-stacks` *-notebook images run the Notebook server as
# user `jovyan`, and set the notebook directory to `/home/jovyan/work`.
# We follow the same convention.
notebook_dir = os.environ.get("DOCKER_NOTEBOOK_DIR", "/home/jovyan/work")
c.DockerSpawner.notebook_dir = notebook_dir

# Mount the real user's Docker volume on the host to the notebook user's
# notebook directory in the container
c.DockerSpawner.volumes = {"jupyterhub-user-{username}": notebook_dir}

# Remove containers once they are stopped
c.DockerSpawner.remove = True

# For debugging arguments passed to spawned containers
c.DockerSpawner.debug = True

# User containers will access hub by container name on the Docker network
c.JupyterHub.hub_ip = "jupyterhub"
c.JupyterHub.hub_port = 8080

# Persist hub data on volume mounted inside container
c.JupyterHub.cookie_secret_file = "/data/jupyterhub_cookie_secret"
c.JupyterHub.db_url = "sqlite:////data/jupyterhub.sqlite"

# Authenticate users with Native Authenticator
c.JupyterHub.authenticator_class = "nativeauthenticator.NativeAuthenticator"

# Allow anyone to sign-up without approval
c.NativeAuthenticator.open_signup = True

# Allowed admins
admin = os.environ.get("JUPYTERHUB_ADMIN")
if admin:
    c.Authenticator.admin_users = [admin]

from jhub_apps.configuration import install_jhub_apps  # noqa: E402
from dockerspawner import DockerSpawner  # noqa: E402

c.DockerSpawner.image = "jhub:latest"
c.JupyterHub.bind_url = "http://0.0.0.0:8000"
c.DockerSpawner.name_template = "{prefix}-{username}-{servername}"
c = install_jhub_apps(c, spawner_to_subclass=DockerSpawner)
c.JupyterHub.log_level = 10



import json
import os

from fastapi import HTTPException, Security, status
from fastapi.security import OAuth2AuthorizationCodeBearer, APIKeyCookie
from fastapi.security.api_key import APIKeyQuery

from .auth import get_jhub_token_from_jwt_token
from .client import get_client
from .models import User

### Endpoints can require authentication using Depends(get_current_user)
### get_current_user will look for a token in url params or
### Authorization: bearer token (header).
### Hub technically supports cookie auth too, but it is deprecated so
### not being included here.
auth_by_param = APIKeyQuery(name="token", auto_error=False)

auth_by_cookie = APIKeyCookie(name="access_token")
auth_url = os.environ["PUBLIC_HOST"] + "/hub/api/oauth2/authorize"
auth_by_header = OAuth2AuthorizationCodeBearer(
    authorizationUrl=auth_url, tokenUrl="oauth_callback", auto_error=False
)
### ^^ The flow for OAuth2 in Swagger is that the "authorize" button
### will redirect user (browser) to "auth_url", which is the Hub login page.
### After logging in, the browser will POST to our internal /get_token endpoint
### with the auth code.  That endpoint POST's to Hub /oauth2/token with
### our client_secret (JUPYTERHUB_API_TOKEN) and that code to get an
### access_token, which it returns to browser, which places in Authorization header.

if os.environ.get("JUPYTERHUB_OAUTH_SCOPES"):
    # typically ["access:services", "access:services!service=$service_name"]
    access_scopes = json.loads(os.environ["JUPYTERHUB_OAUTH_SCOPES"])
else:
    access_scopes = ["access:services"]


### For consideration: optimize performance with a cache instead of
### always hitting the Hub api?
async def get_current_user(
    auth_param: str = Security(auth_by_param),
    auth_header: str = Security(auth_by_header),
    auth_cookie: str = Security(auth_by_cookie)
):
    token = auth_param or auth_header or auth_cookie
    if token is None:
        raise HTTPException(
            status.HTTP_401_UNAUTHORIZED,
            detail="Must login with token parameter or Authorization bearer header",
        )

    token = get_jhub_token_from_jwt_token(token)

    async with get_client() as client:
        endpoint = "/user"
        # normally we auth to Hub API with service api token,
        # but this time auth as the user token to get user model
        headers = {"Authorization": f"Bearer {token}"}
        resp = await client.get(endpoint, headers=headers)
        if resp.is_error:
            raise HTTPException(
                status.HTTP_401_UNAUTHORIZED,
                detail={
                    "msg": "Error getting user info from token",
                    "request_url": str(resp.request.url),
                    "token": token,
                    "response_code": resp.status_code,
                    "hub_response": resp.json(),
                },
            )
    user = User(**resp.json())
    if any(scope in user.scopes for scope in access_scopes):
        return user
    else:
        raise HTTPException(
            status.HTTP_403_FORBIDDEN,
            detail={
                "msg": f"User not authorized: {user.name}",
                "request_url": str(resp.request.url),
                "token": token,
                "user": resp.json(),
            },
        )



import typing
from dataclasses import dataclass
from enum import Enum


@dataclass
class FrameworkConf:
    name: str
    display_name: str
    logo: str


@dataclass
class UserOptions:
    jhub_app: bool
    display_name: str
    description: str
    thumbnail: str
    filepath: str
    framework: str
    custom_command: typing.Optional[str] = None
    conda_env: typing.Optional[dict] = None
    profile: typing.Optional[str] = None


class Framework(Enum):
    panel = "panel"
    bokeh = "bokeh"
    streamlit = "streamlit"
    plotlydash = "plotlydash"
    voila = "voila"
    gradio = "gradio"
    jupyterlab = "jupyterlab"
    custom = "custom"

    @classmethod
    def values(cls):
        return [member.value for role, member in cls.__members__.items()]


FRAMEWORKS = [
    FrameworkConf(
        name=Framework.panel.value,
        display_name="Panel",
        logo="/services/japps/static/img/logos/panel.png",
    ),
    FrameworkConf(
        name=Framework.bokeh.value,
        display_name="Bokeh",
        logo="/services/japps/static/img/logos/bokeh.png",
    ),
    FrameworkConf(
        name=Framework.streamlit.value,
        display_name="Streamlit",
        logo="/services/japps/static/img/logos/streamlit.png",
    ),
    FrameworkConf(
        name=Framework.voila.value,
        display_name="Voila",
        logo="/services/japps/static/img/logos/voila.png",
    ),
    FrameworkConf(
        name=Framework.plotlydash.value,
        display_name="PlotlyDash",
        logo="/services/japps/static/img/logos/plotly-dash.png",
    ),
    FrameworkConf(
        name=Framework.gradio.value,
        display_name="Gradio",
        logo="/services/japps/static/img/logos/gradio.png",
    ),
    # FrameworkConf(
    #     name=Framework.jupyterlab.value,
    #     display_name="JupyterLab",
    #     logo="https://upload.wikimedia.org/wikipedia/commons/thumb/3/38/Jupyter_logo.svg/1200px-Jupyter_logo.svg.png",
    # ),
    FrameworkConf(
        name=Framework.custom.value,
        display_name="Custom Command",
        logo="/services/japps/static/img/logos/jupyter.png",
    ),
]

FRAMEWORKS_MAPPING = {framework.name: framework for framework in FRAMEWORKS}



from base64 import b64encode
from secrets import token_bytes

from traitlets.config import LazyConfigValue

from jhub_apps import JAppsConfig
from jhub_apps.spawner.spawner_creation import subclass_spawner


def _create_token_for_service():
    return b64encode(token_bytes(32)).decode()


def set_defaults_for_jhub_apps_config(c):
    """Set default values for configuration items which are not set by the user"""
    trait_names = JAppsConfig().trait_names()
    trait_names.remove('parent')
    trait_names.remove('log')
    trait_names.remove('config')
    defaults = JAppsConfig().trait_defaults()
    for trait_name in trait_names:
        if isinstance(getattr(c.JAppsConfig, trait_name), LazyConfigValue):
            setattr(c.JAppsConfig, trait_name, defaults.get(trait_name))


def install_jhub_apps(c, spawner_to_subclass):
    c.JupyterHub.spawner_class = subclass_spawner(spawner_to_subclass)
    c.JupyterHub.allow_named_servers = True
    bind_url = c.JupyterHub.bind_url

    set_defaults_for_jhub_apps_config(c)
    if not isinstance(bind_url, str):
        raise ValueError(f"c.JupyterHub.bind_url is not set: {c.JupyterHub.bind_url}")
    if not c.JupyterHub.services:
        c.JupyterHub.services = []
    public_host = c.JupyterHub.bind_url
    fast_api_service_name = "japps"
    oauth_redirect_uri = (
        f"{public_host}/services/{fast_api_service_name}/oauth_callback"
    )
    c.JupyterHub.services.extend(
        [
            {
                "name": fast_api_service_name,
                "url": f"http://{c.JAppsConfig.hub_host}:10202",
                "command": [
                    c.JAppsConfig.python_exec,
                    "-m",
                    "uvicorn",
                    "jhub_apps.service.app:app",
                    "--port=10202",
                    "--host=0.0.0.0",
                    f"--workers={c.JAppsConfig.service_workers}",
                ],
                "environment": {
                    "PUBLIC_HOST": c.JupyterHub.bind_url,
                    "JHUB_APP_TITLE": c.JAppsConfig.app_title,
                    "JHUB_APP_ICON": c.JAppsConfig.app_icon,
                    "JHUB_JUPYTERHUB_CONFIG": c.JAppsConfig.jupyterhub_config_path,
                    "JHUB_APP_JWT_SECRET_KEY": _create_token_for_service(),

                    # Temp environment variables for Nebari Deployment
                    "PROXY_API_SERVICE_PORT": "*",
                    "HUB_SERVICE_PORT": "*",
                },
                "oauth_redirect_uri": oauth_redirect_uri,
                "display": False,
            },
        ]
    )

    services_roles = [
        {
            "name": "japps-service-role",  # name the role
            "services": [
                "japps",  # assign the service to this role
            ],
            "scopes": [
                # declare what permissions the service should have
                "list:users",  # list users
                "read:users:activity",  # read user last-activity
                "read:users",  # read user last-activity
                "admin:servers",  # start/stop servers
                "admin:server_state",  # start/stop servers
                "admin:server_state",  # start/stop servers
                "admin:auth_state",
                "access:services",
                "list:services",
                "read:services",  # read service models
            ],
        },
        {
            "name": "user",
            # grant all users access to services
            "scopes": [
                "self",
                "access:services",
                "admin:auth_state"
            ],
        },
    ]

    if not c.JupyterHub.load_roles:
        c.JupyterHub.load_roles = []

    if isinstance(c.JupyterHub.load_roles, list):
        c.JupyterHub.load_roles = c.JupyterHub.load_roles + services_roles
    else:
        c.JupyterHub.load_roles = services_roles
    return c



from collections import namedtuple
import altair as alt
import math
import pandas as pd
import streamlit as st

"""
# Welcome to Streamlit!

Edit `/streamlit_app.py` to customize this app to your heart's desire :heart:

If you have any questions, checkout our [documentation](https://docs.streamlit.io) and [community
forums](https://discuss.streamlit.io).

In the meantime, below is an example of what you can do with just a few lines of code:
"""


with st.echo(code_location="below"):
    total_points = st.slider("Number of points in spiral", 1, 5000, 2000)
    num_turns = st.slider("Number of turns in spiral", 1, 100, 9)

    Point = namedtuple("Point", "x y")
    data = []

    points_per_turn = total_points / num_turns

    for curr_point_num in range(total_points):
        curr_turn, i = divmod(curr_point_num, points_per_turn)
        angle = (curr_turn + 1) * 2 * math.pi * i / points_per_turn
        radius = curr_point_num / total_points
        x = radius * math.cos(angle)
        y = radius * math.sin(angle)
        data.append(Point(x, y))

    st.altair_chart(
        alt.Chart(pd.DataFrame(data), height=500, width=500)
        .mark_circle(color="#0068c9", opacity=0.5)
        .encode(x="x:Q", y="y:Q")
    )



import logging

from jhub_apps.spawner.utils import get_origin_host
from jhub_apps.spawner.command import (
    EXAMPLES_DIR,
    COMMANDS,
    Command,
    EXAMPLES_FILE,
    DEFAULT_CMD,
    GENERIC_ARGS,
)
from jhub_apps.spawner.types import Framework


logger = logging.getLogger(__name__)


def subclass_spawner(base_spawner):
    # TODO: Find a better way to do this
    class JHubSpawner(base_spawner):

        async def _get_user_auth_state(self):
            try:
                auth_state = await self.user.get_auth_state()
                return auth_state
            except Exception as e:
                logger.exception(e)

        def get_args(self):
            """Return arguments to pass to the notebook server"""
            # logger.info("Getting Spawner args")
            # self._get_user_auth_state()
            argv = super().get_args()
            if self.user_options.get("argv"):
                argv.extend(self.user_options["argv"])

            if self.user_options.get("jhub_app"):
                filepath = self.user_options["filepath"]
                env = self.get_env()
                jh_service_prefix = env.get("JUPYTERHUB_SERVICE_PREFIX")
                framework = self.user_options.get("framework")
                app_filepath = None
                if framework not in [
                    Framework.jupyterlab.value,
                    Framework.custom.value,
                ]:
                    app_filepath = filepath or EXAMPLES_DIR / EXAMPLES_FILE.get(
                        framework
                    )

                if not filepath:
                    # Saving the examples file path when not provided
                    self.user_options["filepath"] = str(app_filepath)

                custom_cmd = self.user_options.get("custom_command")
                if framework == Framework.custom.value:
                    assert custom_cmd
                    command = Command(args=GENERIC_ARGS + custom_cmd.split())
                else:
                    command: Command = COMMANDS.get(framework)
                command_args = command.get_substituted_args(
                    python_exec=self.config.JAppsConfig.python_exec,
                    filepath=app_filepath,
                    origin_host=get_origin_host(self.config.JupyterHub.bind_url),
                    base_url=self.config.JupyterHub.bind_url,
                    jh_service_prefix=jh_service_prefix,
                    voila_base_url=f"{jh_service_prefix}",
                    conda_env=self.user_options.get("conda_env", ""),
                )
                argv.extend(command_args)
            return argv

        def get_env(self):
            # logger.info("Getting spawner environments")
            # await self._get_user_auth_state()
            env = super().get_env()
            if self.user_options.get("env"):
                env.update(self.user_options["env"])

            if self.user_options.get("jhub_app"):
                framework = self.user_options.get("framework")
                jh_service_prefix = env.get("JUPYTERHUB_SERVICE_PREFIX")
                if framework == Framework.plotlydash.value:
                    env["DASH_REQUESTS_PATHNAME_PREFIX"] = jh_service_prefix
                elif framework == Framework.bokeh.value:
                    # Seems like the bokeh is always loading static files
                    # from localhost, this is a bug in bokeh, seen on 3.2.2
                    # at the time of writing this, this environment variable
                    # will load static files from cdn
                    # See this https://github.com/bokeh/bokeh/issues/13170
                    env["BOKEH_RESOURCES"] = "cdn"
            return env

        async def start(self):
            logger.info("Starting spawner process")
            await self._get_user_auth_state()
            framework = self.user_options.get("framework")
            if (
                self.user_options.get("jhub_app")
                and framework != Framework.jupyterlab.value
            ):
                auth_type = "oauth"
                if self.user_options.get("public", False):
                    auth_type = "none"
                self.cmd = DEFAULT_CMD.get_substituted_args(
                    python_exec=self.config.JAppsConfig.python_exec,
                    authtype=auth_type,
                )
            if framework == Framework.jupyterlab.value:
                self.cmd = [
                    self.config.JAppsConfig.python_exec,
                    "-m",
                    "jupyterhub.singleuser",
                ]
            logger.info(f"Final Spawner Command: {self.cmd}")
            return await super().start()

        def _expand_user_vars(self, string):
            """
            Expand user related variables in a given string

            Currently expands:
              {USERNAME} -> Name of the user
              {USERID} -> UserID
              {JHUBSERVERNAME} -> Name
            """
            jhub_server_name = ""
            if self.name:
                jhub_server_name = "-{}".format(self.name)
            return (
                string.replace("{USERNAME}", self.user.name)
                .replace("{USERID}", str(self.user.id))
                .replace("{JHUBSERVERNAME}", jhub_server_name)
            )

    return JHubSpawner



import typing
from datetime import datetime
from typing import Any, Dict, List, Optional


from pydantic import BaseModel


# https://jupyterhub.readthedocs.io/en/stable/_static/rest-api/index.html
class Server(BaseModel):
    name: str
    ready: bool
    pending: Optional[str]
    url: str
    progress_url: str
    started: datetime
    last_activity: datetime
    state: Optional[Any]
    user_options: Optional[Any]


class User(BaseModel):
    name: str
    admin: bool
    groups: Optional[List[str]]
    kind: str
    server: Optional[str] = None
    pending: Optional[str] = None
    last_activity: Optional[datetime] = None
    servers: Optional[Dict[str, Server]] = None
    scopes: List[str]
    auth_state: Optional[Dict] = None


# https://stackoverflow.com/questions/64501193/fastapi-how-to-use-httpexception-in-responses
class AuthorizationError(BaseModel):
    detail: str


class HubResponse(BaseModel):
    msg: str
    request_url: str
    token: str
    response_code: int
    hub_response: dict


class HubApiError(BaseModel):
    detail: HubResponse


class UserOptions(BaseModel):
    jhub_app: bool
    display_name: str
    description: str
    thumbnail: str = None
    filepath: typing.Optional[str] = str()
    framework: str = "panel"
    custom_command: typing.Optional[str] = str()
    conda_env: typing.Optional[str] = str()
    profile: typing.Optional[str] = str()
    # Make app available to public (unauthenticated Hub users)
    public: typing.Optional[bool] = False


class ServerCreation(BaseModel):
    servername: str
    user_options: UserOptions



import argparse
import gradio as gr


def greet(name):
    return "Hello " + name + "!"


demo = gr.Interface(fn=greet, inputs="text", outputs="text")


parser = argparse.ArgumentParser(description="Process CLI args for gradio")
parser.add_argument(
    "--server-port", type=str, help="server_port for gradio app", default=8500
)
parser.add_argument("--root-path", type=str, help="root_path for gradio", default=None)
cli_args = parser.parse_args()

demo.launch(server_port=int(cli_args.server_port), root_path=cli_args.root_path)



import string
import typing
from dataclasses import dataclass
from pathlib import Path

from jhub_apps.spawner.types import Framework


EXAMPLES_FILE = {
    Framework.panel.value: "panel_basic.py",
    Framework.bokeh.value: "bokeh_basic.py",
    Framework.streamlit.value: "streamlit_app.py",
    Framework.plotlydash.value: "plotlydash_app.py",
    Framework.voila.value: "voila_basic.ipynb",
    Framework.gradio.value: "gradio_basic.py",
}

EXAMPLES_DIR = Path(__file__).parent.parent / "examples"


@dataclass
class TString:
    value: str

    def replace(self, **kwargs):
        template = string.Template(self.value)
        keys_to_substitute = set()
        for k, v in kwargs.items():
            if f"${k}" in self.value:
                keys_to_substitute.add(k)
        subs = {k: v for k, v in kwargs.items() if k in keys_to_substitute}
        return template.substitute(subs)


@dataclass
class Command:
    args: typing.List[str]

    def get_substituted_args(self, **kwargs):
        subs_args = []
        for arg in self.args:
            s_arg = arg
            if isinstance(arg, TString):
                s_arg = arg.replace(**kwargs)
            subs_args.append(s_arg)
        return subs_args


DEFAULT_CMD = Command(
    args=[
        TString("$python_exec"),
        "-m",
        "jhsingle_native_proxy.main",
        TString("--authtype=$authtype"),
    ]
)

GENERIC_ARGS = [
    "--destport=0",
    "--ready-check-path=/ready-check",
    TString("--conda-env=$conda_env"),
    TString("$python_exec"),
    "{-}m",
]

COMMANDS = {
    Framework.gradio.value: Command(
        args=[
            "--destport=0",
            TString("--conda-env=$conda_env"),
            TString("$python_exec"),
            TString("$filepath"),
            "{--}server-port={port}",
            TString("{--}root-path=$jh_service_prefix"),
            "--ready-check-path=/",
        ],
    ),
    Framework.voila.value: Command(
        args=[
            "--destport=0",
            TString("--conda-env=$conda_env"),
            TString("$python_exec"),
            "{-}m",
            "voila",
            TString("$filepath"),
            "{--}port={port}",
            "{--}no-browser",
            "{--}Voila.server_url=/",
            "{--}Voila.ip=0.0.0.0",
            "{--}Voila.tornado_settings",
            "--debug",
            TString("allow_origin=$origin_host"),
            TString("{--}Voila.base_url=$voila_base_url"),
            "--progressive",
            "--ready-check-path=/voila/static/",
        ],
    ),
    Framework.streamlit.value: Command(
        args=[
            "--destport=0",
            TString("--conda-env=$conda_env"),
            TString("$python_exec"),
            "{-}m",
            "streamlit",
            "run",
            TString("$filepath"),
            "{--}server.port={port}",
            "{--}server.headless=True",
            TString("{--}browser.serverAddress=$origin_host"),
            "{--}browser.gatherUsageStats=false",
        ],
    ),
    Framework.plotlydash.value: Command(
        args=[
            "--destport=0",
            TString("--conda-env=$conda_env"),
            TString("$python_exec"),
            "{-}m",
            "plotlydash_tornado_cmd.main",
            TString("$filepath"),
            "{--}port={port}",
        ],
    ),
    Framework.bokeh.value: Command(
        args=[
            "--destport=0",
            TString("--conda-env=$conda_env"),
            TString("$python_exec"),
            "{-}m",
            "bokeh_root_cmd.main",
            TString("$filepath"),
            "{--}port={port}",
            TString("{--}allow-websocket-origin=$origin_host"),
            TString("{--}prefix=$base_url"),
            "--ip=0.0.0.0",
            "--ready-check-path=/ready-check",
        ]
    ),
    Framework.panel.value: Command(
        args=[
            "--destport=0",
            TString("--conda-env=$conda_env"),
            TString("$python_exec"),
            "{-}m",
            "bokeh_root_cmd.main",
            TString("$filepath"),
            "{--}port={port}",
            "{--}debug",
            TString("{--}allow-websocket-origin=$origin_host"),
            "{--}server=panel",
            TString("{--}prefix=$base_url"),
            "--ip=0.0.0.0",
            "--ready-check-path=/ready-check",
        ]
    ),
    Framework.jupyterlab.value: Command(args=[]),
}



import logging
import os
import re
import uuid

import requests


API_URL = os.environ.get("JUPYTERHUB_API_URL")
JUPYTERHUB_API_TOKEN = os.environ.get("JUPYTERHUB_API_TOKEN")

logger = logging.getLogger(__name__)


class HubClient:
    def __init__(self, token=None):
        self.token = token or JUPYTERHUB_API_TOKEN

    def _headers(self):
        return {"Authorization": f"token {self.token}"}

    def get_users(self):
        r = requests.get(
            API_URL + "/users",
            params={"include_stopped_servers": True},
            headers=self._headers()
        )
        r.raise_for_status()
        users = r.json()
        return users

    def get_user(self, user):
        r = requests.get(
            API_URL + f"/users/{user}",
            params={"include_stopped_servers": True},
            headers=self._headers()
        )
        r.raise_for_status()
        user = r.json()
        return user

    def get_server(self, username, servername):
        user = self.get_user(username)
        for name, server in user["servers"].items():
            if name == servername:
                return server

    def normalize_server_name(self, servername):
        # Convert text to lowercase
        text = servername.lower()
        # Remove all special characters except spaces and hyphen
        text = re.sub(r"[^a-z0-9\s-]", "", text)
        # Replace spaces with hyphens
        text = text.replace(" ", "-")
        return text

    def start_server(self, username, servername):
        if not servername:
            logger.info("Starting JupyterLab server")
            # Default server, which is JupyterLab (not named server)
            servername = ""
            user_options = {}
        else:
            # Get named server
            server = self.get_server(username, servername)
            user_options = server["user_options"]
            if not server:
                return None
        url = f"/users/{username}/servers/{servername}"
        data = {"name": servername, **user_options}
        r = requests.post(API_URL + url, headers=self._headers(), json=data)
        r.raise_for_status()
        return r.status_code, servername

    def create_server(self, username, servername, user_options=None):
        logger.info("Creating new server")
        servername = self.normalize_server_name(servername)
        servername = f"{servername}-{uuid.uuid4().hex[:7]}"
        return self._create_server(username, servername, user_options)

    def edit_server(self, username, servername, user_options=None):
        logger.info(f"Editing server: {servername}")
        server = self.get_server(username, servername)
        if server:
            # Stop the server first
            logger.info(f"Stopping the server: {servername} first")
            self.delete_server(username, server["name"])
        else:
            raise ValueError("Server does not exists")
        logger.info(f"Now creating the server: {servername} with new params")
        return self._create_server(username, servername, user_options)

    def _create_server(self, username, servername, user_options):
        url = f"/users/{username}/servers/{servername}"
        params = user_options.model_dump()
        data = {"name": servername, **params}
        logger.info("Creating new server")
        r = requests.post(API_URL + url, headers=self._headers(), json=data)
        r.raise_for_status()
        return r.status_code, servername

    def delete_server(self, username, server_name, remove=False):
        if server_name is None:
            # Default server and not named server
            server_name = ""
        url = f"/users/{username}/servers/{server_name}"
        # This will remove it from the database, otherwise it will just stop the server
        params = {"remove": remove}
        r = requests.delete(API_URL + url, headers=self._headers(), json=params)
        r.raise_for_status()
        return r.status_code

    def get_services(self):
        r = requests.get(API_URL + "/services", headers=self._headers())
        r.raise_for_status()
        return r.json()



import argparse

from jhub_apps.version import get_version


def app():
    parser = argparse.ArgumentParser(description="Get the version of a package")
    parser.add_argument("--version", action="store_true", help="Print the version of jhub-apps")
    args = parser.parse_args()
    if args.version:
        version_info = get_version()
        print(version_info)



from bokeh.models import ColumnDataSource
from bokeh.plotting import figure
from bokeh.io import curdoc


def modify_doc(doc):
    """Add a plotted function to the document.

    Arguments:
        doc: A bokeh document to which elements can be added.
    """
    x_values = list(range(10))
    y_values = [x**2 for x in x_values]
    data_source = ColumnDataSource(data=dict(x=x_values, y=y_values))
    plot = figure(
        title="f(x) = x^2",
        tools="crosshair,pan,reset,save,wheel_zoom",
    )
    plot.line("x", "y", source=data_source, line_width=3, line_alpha=0.6)
    doc.add_root(plot)
    doc.title = "Hello World"


def main():
    modify_doc(curdoc())


main()



LOGO = "/services/japps/static/img/Nebari-Logo-Horizontal-Lockup-White-text.svg"

DEFAULT_THEME = {
    "logo": LOGO,
    "primary_color": "#C316E9",
    "primary_color_dark": "#79158a",
    "secondary_color": "#18817A",
    "secondary_color_dark": "#12635E",
    "accent_color": "#eda61d",
    "accent_color_dark": "#a16d14",
    "text_color": "#1c1d26",
    "h1_color": "#0f1015",
    "h2_color": "#0f1015",
    "navbar_text_color": "#ffffff",
    "navbar_hover_color": "#20b1a8",
    "navbar_color": "#1c1d26",
}



import logging
import os
import typing
from datetime import timedelta, datetime

import jwt
from fastapi import HTTPException, status

logger = logging.getLogger(__name__)


def create_access_token(data: dict, expires_delta: typing.Optional[timedelta] = None):
    logger.info("Creating access token")
    to_encode = data.copy()
    if expires_delta:
        expire = datetime.utcnow() + expires_delta
    else:
        expire = datetime.utcnow() + timedelta(minutes=15)
    to_encode.update({"exp": expire})
    secret_key = os.environ["JHUB_APP_JWT_SECRET_KEY"]
    encoded_jwt = jwt.encode(to_encode, secret_key, algorithm="HS256")
    return encoded_jwt


def get_jhub_token_from_jwt_token(token):
    logger.info("Trying to get JHub Apps token from JWT Token")
    credentials_exception = HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail={
            "msg": "Could not validate credentials"
        },
        headers={"WWW-Authenticate": "Bearer"},
    )
    try:
        payload = jwt.decode(token, os.environ["JHUB_APP_JWT_SECRET_KEY"], algorithms=["HS256"])
        access_token_data: dict = payload.get("sub")
        if access_token_data is None:
            raise credentials_exception
    except jwt.PyJWTError as e:
        logger.warning("Authentication failed for token")
        logger.exception(e)
        raise credentials_exception
    logger.info("Fetched access token from JWT Token")
    return access_token_data["access_token"]



import logging
import os
from pathlib import Path

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles

from .routes import router

### When managed by Jupyterhub, the actual endpoints
### will be served out prefixed by /services/:name.
### One way to handle this with FastAPI is to use an APIRouter.
### All routes are defined in service.py

STATIC_DIR = Path(__file__).parent.parent / "static"


def setup_logging():
    # TODO: Use structlog
    logging_format = (
        "%(asctime)s %(levelname)9s %(lineno)4s %(module)s: %(message)s"
    )
    logging.basicConfig(
        level=logging.INFO, format=logging_format
    )


app = FastAPI(
    title="JApps Service",
    version="0.1",
    ### Serve out Swagger from the service prefix (<hub>/services/:name/docs)
    openapi_url=router.prefix + "/openapi.json",
    docs_url=router.prefix + "/docs",
    redoc_url=router.prefix + "/redoc",
    ### Add our service client id to the /docs Authorize form automatically
    swagger_ui_init_oauth={"clientId": os.environ["JUPYTERHUB_CLIENT_ID"]},
    swagger_ui_parameters={"persistAuthorization": True},
    ### Default /docs/oauth2 redirect will cause Hub
    ### to raise oauth2 redirect uri mismatch errors
    # swagger_ui_oauth2_redirect_url=os.environ["JUPYTERHUB_OAUTH_CALLBACK_URL"],
)
static_files = StaticFiles(directory=STATIC_DIR)
app.mount(f"{router.prefix}/static", static_files, name="static")
app.include_router(router)
setup_logging()



import logging
import typing
import dataclasses
import os
from datetime import timedelta

import requests
from fastapi import APIRouter, Depends, status, Request, File, UploadFile, Form, HTTPException
from fastapi.encoders import jsonable_encoder
from pydantic import BaseModel, ValidationError
from starlette.responses import RedirectResponse

from jhub_apps.service.auth import create_access_token
from jhub_apps.service.client import get_client
from jhub_apps.service.models import AuthorizationError, HubApiError, User, ServerCreation
from jhub_apps.service.security import get_current_user

from fastapi import FastAPI
from fastapi.templating import Jinja2Templates

from jhub_apps.hub_client.hub_client import HubClient
from jhub_apps.service.utils import get_conda_envs, get_jupyterhub_config, get_spawner_profiles, get_thumbnail_data_url
from jhub_apps.spawner.types import FRAMEWORKS
from jhub_apps.version import get_version

app = FastAPI()

logger = logging.getLogger(__name__)
templates = Jinja2Templates(directory="jhub_apps/templates")

# Expires in 7 days
ACCESS_TOKEN_EXPIRE_MINUTES = 60 * 24 * 7

# APIRouter prefix cannot end in /
service_prefix = os.getenv("JUPYTERHUB_SERVICE_PREFIX", "").rstrip("/")
router = APIRouter(prefix=service_prefix)

# TODO: Add response models for all endpoints


@router.get("/oauth_callback", include_in_schema=False)
async def get_token(code: str):
    "Callback function for OAuth2AuthorizationCodeBearer scheme"
    # The only thing we need in this form post is the code
    # Everything else we can hardcode / pull from env
    logger.info(f"Getting token for code {code}")
    async with get_client() as client:
        redirect_uri = (
            os.environ["PUBLIC_HOST"] + os.environ["JUPYTERHUB_OAUTH_CALLBACK_URL"],
        )
        data = {
            "client_id": os.environ["JUPYTERHUB_CLIENT_ID"],
            "client_secret": os.environ["JUPYTERHUB_API_TOKEN"],
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": redirect_uri,
        }
        resp = await client.post("/oauth2/token", data=data)
    access_token_expires = timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES)
    access_token = create_access_token(
        data={"sub": resp.json()}, expires_delta=access_token_expires
    )
    ### resp.json() is {'access_token': <token>, 'token_type': 'Bearer'}
    response = RedirectResponse(os.environ["PUBLIC_HOST"] + "/hub/home", status_code=302)
    response.set_cookie(key="access_token", value=access_token, httponly=True)
    return response


@router.get("/jhub-login", description="Login via OAuth2")
async def login(request: Request):
    logger.info(f"Logging in: {request}")
    authorization_url = os.environ["PUBLIC_HOST"] + "/hub/api/oauth2/authorize?response_type=code&client_id=service-japps"
    return RedirectResponse(authorization_url, status_code=302)


def get_all_shared_servers(hub_users, current_hub_user):
    all_servers = []
    for hub_user in hub_users:
        if hub_user['name'] != current_hub_user['name']:
            hub_user_servers = list(hub_user["servers"].values())
            hub_user_servers_with_name = [{"username": hub_user['name'], **server} for server in hub_user_servers]
            all_servers.extend(hub_user_servers_with_name)
    # Filter default servers
    return list(filter(lambda server: server['name'] != '', all_servers))


@router.get("/server/", description="Get all servers")
@router.get("/server/{server_name}", description="Get a server by server name")
async def get_server(user: User = Depends(get_current_user), server_name=None):
    """Get servers for the authenticated user"""
    hub_client = HubClient()
    users = hub_client.get_users()
    current_hub_user = None
    for hub_user in users:
        if hub_user['name'] == user.name:
            current_hub_user = hub_user
            break
    if not current_hub_user:
        raise HTTPException(
            detail=f"Hub user '{user.name}' not found",
            status_code=status.HTTP_404_NOT_FOUND,
        )
    user_servers = current_hub_user["servers"]
    if server_name:
        # Get a particular server
        for s_name, server_details in user_servers.items():
            if s_name == server_name:
                return server_details
        raise HTTPException(
            detail=f"server '{server_name}' not found",
            status_code=status.HTTP_404_NOT_FOUND,
        )
    else:
        # Get all servers
        return {
            "shared_apps": get_all_shared_servers(hub_users=users, current_hub_user=current_hub_user),
            "user_apps": list(user_servers.values())
        }


class Checker:
    def __init__(self, model: BaseModel):
        self.model = model

    def __call__(self, data: str = Form(...)):
        try:
            return self.model.model_validate_json(data)
        except ValidationError as e:
            raise HTTPException(
                detail=jsonable_encoder(e.errors()),
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            )


@router.post("/server")
async def create_server(
    server: ServerCreation = Depends(Checker(ServerCreation)),
    thumbnail: typing.Optional[UploadFile] = File(None),
    user: User = Depends(get_current_user),
):
    server.user_options.thumbnail = await get_thumbnail_data_url(
        framework_name=server.user_options.framework,
        thumbnail=thumbnail
    )
    hub_client = HubClient()
    return hub_client.create_server(
        username=user.name,
        servername=server.servername,
        user_options=server.user_options,
    )


@router.post("/server/")
@router.post("/server/{server_name}")
async def start_server(
        server_name=None,
        user: User = Depends(get_current_user),
):
    """Start an already existing server."""
    hub_client = HubClient()
    try:
        response = hub_client.start_server(
            username=user.name,
            servername=server_name,
        )
    except requests.exceptions.HTTPError as e:
        raise HTTPException(
            detail=f"Probably server '{server_name}' is already running: {e}",
            status_code=status.HTTP_400_BAD_REQUEST,
        )
    if response is None:
        raise HTTPException(
            detail=f"server '{server_name}' not found",
            status_code=status.HTTP_404_NOT_FOUND,
        )
    return response


@router.put("/server/{server_name}")
async def update_server(
    server: ServerCreation = Depends(Checker(ServerCreation)),
    thumbnail: typing.Optional[UploadFile] = File(None),
    user: User = Depends(get_current_user), server_name=None
):
    server.user_options.thumbnail = await get_thumbnail_data_url(
        framework_name=server.user_options.framework,
        thumbnail=thumbnail
    )
    hub_client = HubClient()
    return hub_client.edit_server(
        username=user.name,
        servername=server_name,
        user_options=server.user_options,
    )


@router.delete("/server/{server_name}")
@router.delete("/server/")
async def delete_server(
        user: User = Depends(get_current_user),
        server_name=None,
        remove: bool = False,
):
    """Delete or stop server. Delete if remove is True otherwise stop the server"""
    hub_client = HubClient()
    return hub_client.delete_server(
        user.name,
        server_name=server_name,
        remove=remove
    )


@router.get(
    "/user",
    response_model=User,
    responses={401: {"model": AuthorizationError}, 400: {"model": HubApiError}},
)
async def me(user: User = Depends(get_current_user)):
    """Authenticated function that returns the User model"""
    return user


@router.get("/frameworks/", description="Get all frameworks")
async def get_frameworks(user: User = Depends(get_current_user)):
    frameworks = []
    for framework in FRAMEWORKS:
        frameworks.append(dataclasses.asdict(framework))
    return frameworks


@router.get("/conda-environments/", description="Get all conda environments")
async def conda_environments(user: User = Depends(get_current_user)):
    logging.info(f"Getting conda environments for user: {user}")
    config = get_jupyterhub_config()
    hclient = HubClient()
    user_from_service = hclient.get_user(user.name)
    conda_envs = get_conda_envs(config, user_from_service)
    logger.info(f"Found conda environments: {conda_envs}")
    return conda_envs


@router.get("/spawner-profiles/", description="Get all spawner profiles")
async def spawner_profiles(user: User = Depends(get_current_user)):
    hclient = HubClient()
    user_from_service = hclient.get_user(user.name)
    auth_state = user_from_service.get("auth_state")
    logging.info(f"Getting spawner profiles for user: {user.name}")
    config = get_jupyterhub_config()
    spawner_profiles_ = await get_spawner_profiles(config, auth_state=auth_state)
    logger.info(f"Loaded spawner profiles: {spawner_profiles_}")
    return spawner_profiles_


@router.get("/services/", description="Get all services")
async def hub_services(user: User = Depends(get_current_user)):
    logging.info(f"Getting hub services for user: {user}")
    hub_client = HubClient()
    return hub_client.get_services()


@router.get(
    "/status",
)
async def status_endpoint():
    """Check API Status"""
    # TODO: Add version
    version = get_version()
    return {
        "status": "ok",
        "version": str(version)
    }



import base64
import logging
import os
from unittest.mock import Mock

import requests
from jupyterhub.app import JupyterHub
from traitlets.config import LazyConfigValue

from jhub_apps.spawner.types import FrameworkConf, FRAMEWORKS_MAPPING
from slugify import slugify


logger = logging.getLogger(__name__)


def get_jupyterhub_config():
    hub = JupyterHub()
    jhub_config_file_path = os.environ["JHUB_JUPYTERHUB_CONFIG"]
    logger.info(f"Getting JHub config from file: {jhub_config_file_path}")
    hub.load_config_file(jhub_config_file_path)
    config = hub.config
    logger.info(f"JApps config: {config.JAppsConfig}")
    return config


def get_conda_envs(config, user):
    """This will extract conda environment from the JupyterHub config"""
    if isinstance(config.JAppsConfig.conda_envs, list):
        return config.JAppsConfig.conda_envs
    elif isinstance(config.JAppsConfig.conda_envs, LazyConfigValue):
        return []
    elif callable(config.JAppsConfig.conda_envs):
        try:
            logger.info("JAppsConfig.conda_envs is a callable, calling now..")
            return config.JAppsConfig.conda_envs(user)
        except Exception as e:
            logger.exception(e)
            return []
    else:
        raise ValueError(
            f"Invalid value for config.JAppsConfig.conda_envs: {config.JAppsConfig.conda_envs}"
        )


def get_fake_spawner_object(auth_state):
    fake_spawner = Mock()

    async def get_auth_state():
        return auth_state

    fake_spawner.user.get_auth_state = get_auth_state
    fake_spawner.log = logger
    return fake_spawner


def _slugify_profile_list(profile_list):
    # This is replicating the following:
    # https://github.com/jupyterhub/kubespawner/blob/a4b9b190f0335406c33c6de11b5d1b687842dd89/kubespawner/spawner.py#L3279
    # Since we are not inside spawner yet, the profiles might not be slugified yet
    if not profile_list:
        # empty profile lists are just returned
        return profile_list

    for profile in profile_list:
        # generate missing slug fields from display_name
        if 'slug' not in profile:
            profile['slug'] = slugify(profile['display_name'])
    return profile_list


async def get_spawner_profiles(config, auth_state=None):
    """This will extract spawner profiles from the JupyterHub config
    If the Spawner is KubeSpawner
    # See: https://jupyterhub-kubespawner.readthedocs.io/en/latest/spawner.html#kubespawner.KubeSpawner.profile_list
    """
    profile_list = config.KubeSpawner.profile_list
    if isinstance(profile_list, list):
        return config.KubeSpawner.profile_list
    elif isinstance(profile_list, LazyConfigValue):
        return []
    elif callable(profile_list):
        try:
            logger.info("config.KubeSpawner.profile_list is a callable, calling now..")
            profile_list = await profile_list(get_fake_spawner_object(auth_state))
            return _slugify_profile_list(profile_list)
        except Exception as e:
            logger.exception(e)
            return []
    else:
        raise ValueError(
            f"Invalid value for config.KubeSpawner.profile_list: {profile_list}"
        )


def encode_file_to_data_url(filename, file_contents):
    """Converts image file to data url to display in browser."""
    base64_encoded = base64.b64encode(file_contents)
    filename_ = filename.lower()
    if filename_.endswith(".jpg") or filename_.endswith(".jpeg"):
        mime_type = "image/jpeg"
    elif filename_.endswith('.svg'):
        mime_type = "image/svg+xml"
    else:
        file_extension = filename_.split('.')[-1]
        mime_type = f"image/{file_extension}"
    data_url = f"data:{mime_type};base64,{base64_encoded.decode('utf-8')}"
    return data_url


def get_default_thumbnail(framework_name):
    framework: FrameworkConf = FRAMEWORKS_MAPPING.get(framework_name)
    thumbnail_url = framework.logo
    if thumbnail_url.startswith("/"):
        base_url = os.environ["PUBLIC_HOST"]
        thumbnail_url = f"{base_url}{thumbnail_url}"
    try:
        response = requests.get(thumbnail_url)
    except Exception as e:
        logger.info(f"Unable to fetch thumbnail from url: {thumbnail_url}:")
        logger.exception(e)
        return
    if response.status_code == 200:
        thumbnail_content = response.content
        thumbnail_filename = thumbnail_url.split("/")[-1]
        return encode_file_to_data_url(filename=thumbnail_filename, file_contents=thumbnail_content)


async def get_thumbnail_data_url(framework_name, thumbnail):
    if thumbnail:
        thumbnail_contents = await thumbnail.read()
        thumbnail_data_url = encode_file_to_data_url(
            thumbnail.filename, thumbnail_contents
        )
    else:
        thumbnail_data_url = get_default_thumbnail(framework_name)
    return thumbnail_data_url



from dash import Dash, html, dcc, callback, Output, Input
import plotly.express as px
import pandas as pd

df = pd.read_csv(
    "https://raw.githubusercontent.com/plotly/datasets/master/gapminder_unfiltered.csv"
)

app = Dash(__name__)

app.layout = html.Div(
    [
        html.H1(children="Plotly Dash App", style={"textAlign": "center"}),
        dcc.Dropdown(df.country.unique(), "Canada", id="dropdown-selection"),
        dcc.Graph(id="graph-content"),
    ]
)


@callback(Output("graph-content", "figure"), Input("dropdown-selection", "value"))
def update_graph(value):
    dff = df[df.country == value]
    return px.line(dff, x="year", y="pop")


if __name__ == "__main__":
    app.run(debug=True)



from jupyterhub.spawner import SimpleLocalProcessSpawner

from jhub_apps import theme_template_paths, themes
from jhub_apps.configuration import install_jhub_apps

c = get_config()  # noqa

from jupyterhub.auth import DummyAuthenticator  # noqa: E402

c.Authenticator.admin_users = {"admin"}
c.JupyterHub.authenticator_class = DummyAuthenticator
c.JupyterHub.log_level = 10

hub_url = "http://127.0.0.1:8000"
c.JupyterHub.bind_url = hub_url
c.JAppsConfig.jupyterhub_config_path = "jupyterhub_config.py"
c.JAppsConfig.conda_envs = []
c.JAppsConfig.service_workers = 2
c.JupyterHub.default_url = "/hub/home"

c = install_jhub_apps(c, spawner_to_subclass=SimpleLocalProcessSpawner)

c.JupyterHub.template_paths = theme_template_paths


def service_for_jhub_apps(name, url):
    return {
        "name": name,
        "display": True,
        "info": {
            "name": name,
            "url": url,
            "external": True,
        },
        "oauth_no_confirm": True,
    }


c.JupyterHub.services.extend(
    [
        service_for_jhub_apps(name="JupyterLab", url="/user/[USER]/lab"),
        service_for_jhub_apps(name="Argo", url="/argo"),
        service_for_jhub_apps(name="Users", url="/auth/admin/nebari/console/"),
        service_for_jhub_apps(name="Environments", url="/conda-store"),
        service_for_jhub_apps(name="Monitoring", url="/monitoring"),
    ]
)

# nebari will control these as ways to customize the template
c.JupyterHub.template_vars = {
    "hub_title": "Welcome to Nebari",
    "hub_subtitle": "your open source data science platform",
    "welcome": "Running in dev mode",
    **themes.DEFAULT_THEME
}
